package ru.dbotthepony.kstarbound.io.json.util

import com.google.gson.Gson
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.Starbound
import java.util.concurrent.ConcurrentHashMap

/**
 * Данный [TypeAdapter] реализует возможности чтения данных по "ссылке" на файл.
 *
 * К примеру, если в оригинальной структуре может (или обычно) встречаться такое:
 *
 * ```json
 * {
 * 		"thing": { ... },
 * }
 * ```
 *
 * и такое:
 *
 * ```json
 * {
 * 		"thing": "/path/to/thing.config"
 * }
 * ```
 *
 * То второй вариант будет прочитан из указанного пути и закеширован внутри [ReferenceAdapter]
 */
class ReferenceAdapter<T>(val parent: TypeAdapter<T>) : TypeAdapter<T>() {
	private val cache = ConcurrentHashMap<String, T>()

	override fun write(out: JsonWriter, value: T) {
		parent.write(out, value)
	}

	override fun read(reader: JsonReader): T {
		if (reader.peek() != JsonToken.STRING) {
			return parent.read(reader)
		}

		var path = reader.nextString()

		if (path[0] != '/') {
			// относительный путь

			val readingFolder = Starbound.assetFolder ?: throw NullPointerException("Currently read folder is not specified")
			path = "$readingFolder/$path"
		}

		return cache.computeIfAbsent(path) {
			return@computeIfAbsent parent.read(JsonReader(Starbound.locate(it).reader()).also { it.isLenient = true })
		}
	}
}

class ReferenceAdapterFactory(val parent: TypeAdapterFactory) : TypeAdapterFactory {
	override fun <T : Any?> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
		val candidate = parent.create(gson, type)

		if (candidate != null) {
			return ReferenceAdapter(candidate)
		}

		return null
	}
}

fun <T> TypeAdapter<T>.asReference(): TypeAdapter<T> {
	return ReferenceAdapter(this)
}

fun TypeAdapterFactory.asReference(): TypeAdapterFactory {
	return ReferenceAdapterFactory(this)
}
